module  IIR_Filter_p16 #(
 parameter  coef_width          = 32
,parameter  b_pow16_width	= 29
,parameter  ab_pow_width	= 32 
,parameter  data_in_width       = 16 
,parameter  data_out_width	= 16
,parameter  temp_var_width	= 29
)
// H(z) = a(1 + b*z^-1 + b^2*z^-2 + b^3*z^-3 + b^4*z^-4 + b^5*z^-5 + b^6*z^-6 + b^7*z^-7) / (1 - b^8*z^-8)
(
 input	 rstn
,input	 clk
,input	 en
,input   signed	[data_in_width-1 :0]   dinp0	//x(8n+16)
,input   signed	[data_in_width-1 :0]   dinp1 //x(8n+15)
,input   signed	[data_in_width-1 :0]   dinp2 //x(8n+14)
,input   signed	[data_in_width-1 :0]   dinp3 //x(8n+13)
,input   signed	[data_in_width-1 :0]   dinp4 //x(8n+12)
,input   signed	[data_in_width-1 :0]   dinp5 //x(8n+11)
,input   signed	[data_in_width-1 :0]   dinp6 //x(8n+10)
,input   signed	[data_in_width-1 :0]   dinp7 //x(8n+9)
,input   signed	[data_in_width-1 :0]   dinp8
,input   signed	[data_in_width-1 :0]   dinp9 
,input   signed	[data_in_width-1 :0]   dinpa 
,input   signed	[data_in_width-1 :0]   dinpb 
,input   signed	[data_in_width-1 :0]   dinpc 
,input   signed	[data_in_width-1 :0]   dinpd 
,input   signed	[data_in_width-1 :0]   dinpe 
,input   signed	[data_in_width-1 :0]   dinpf 

,input   signed	[coef_width-1    :0]   a_re
,input   signed	[coef_width-1    :0]   ab_re
,input   signed	[coef_width-1    :0]   abb_re
,input   signed	[coef_width-1    :0]   ab_pow3_re
,input   signed	[coef_width-1    :0]   ab_pow4_re
,input   signed	[coef_width-1    :0]   ab_pow5_re
,input   signed	[coef_width-1    :0]   ab_pow6_re
,input   signed	[coef_width-1    :0]   ab_pow7_re
,input   signed	[coef_width-1    :0]   ab_pow8_re
,input   signed	[coef_width-1    :0]   ab_pow9_re
,input   signed	[coef_width-1    :0]   ab_powa_re
,input   signed	[coef_width-1    :0]   ab_powb_re
,input   signed	[coef_width-1    :0]   ab_powc_re
,input   signed	[coef_width-1    :0]   ab_powd_re
,input   signed	[coef_width-1    :0]   ab_powe_re
,input   signed	[coef_width-1    :0]   ab_powf_re
,input   signed	[coef_width-1    :0]   b_pow16_re
,output  signed	[data_out_width-1:0]   dout_re		// Re(y(8n-8))

`ifdef COMPLEX
input   signed	[coef_width-1    :0]   a_im;
input   signed	[coef_width-1    :0]   ab_im;
input   signed	[coef_width-1    :0]   abb_im;
input   signed	[coef_width-1    :0]   ab_pow3_im;
input   signed	[coef_width-1    :0]   ab_pow4_im;
input   signed	[coef_width-1    :0]   ab_pow5_im;
input   signed	[coef_width-1    :0]   ab_pow6_im;
input   signed	[coef_width-1    :0]   ab_pow7_im;
input   signed	[coef_width-1    :0]   ab_pow8_im;
input   signed	[coef_width-1    :0]   ab_pow9_im;
input   signed	[coef_width-1    :0]   ab_powa_im;
input   signed	[coef_width-1    :0]   ab_powb_im;
input   signed	[coef_width-1    :0]   ab_powc_im;
input   signed	[coef_width-1    :0]   ab_powd_im;
input   signed	[coef_width-1    :0]   ab_powe_im;
input   signed	[coef_width-1    :0]   ab_powf_im;
input   signed	[coef_width-1    :0]   b_pow16_im;
output  signed	[data_out_width-1:0]   dout_im;		// Im(y(8n-8))
`endif
);

wire   signed	[data_in_width-1   :0]   dinp  [15:0];
assign dinp[15] = dinpf;
assign dinp[14] = dinpe;
assign dinp[13] = dinpd;
assign dinp[12] = dinpc;
assign dinp[11] = dinpb;
assign dinp[10] = dinpa;
assign dinp[9 ] = dinp9;
assign dinp[8 ] = dinp8;
assign dinp[7 ] = dinp7;
assign dinp[6 ] = dinp6;
assign dinp[5 ] = dinp5;
assign dinp[4 ] = dinp4;
assign dinp[3 ] = dinp3;
assign dinp[2 ] = dinp2;
assign dinp[1 ] = dinp1;
assign dinp[0 ] = dinp0;

wire   signed	[ab_pow_width-1   :0]   ab_pow_re  [15:0];
`ifdef COMPLEX
wire   signed	[ab_pow_width-1   :0]   ab_pow_im  [15:0];
`endif 

assign ab_pow_re[15] = ab_powf_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[14] = ab_powe_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[13] = ab_powd_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[12] = ab_powc_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[11] = ab_powb_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[10] = ab_powa_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[9 ] = ab_pow9_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[8 ] = ab_pow8_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[7 ] = ab_pow7_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[6 ] = ab_pow6_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[5 ] = ab_pow5_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[4 ] = ab_pow4_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[3 ] = ab_pow3_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[2 ] = abb_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[1 ] = ab_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
assign ab_pow_re[0 ] = a_re[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_re[coef_width-ab_pow_width-1];
`ifdef COMPLEX
assign ab_pow_im[15] = ab_powf_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[14] = ab_powe_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[13] = ab_powd_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[12] = ab_powc_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[11] = ab_powb_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[10] = ab_powa_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[9 ] = ab_pow9_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[8 ] = ab_pow8_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[7 ] = ab_pow7_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[6 ] = ab_pow6_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[5 ] = ab_pow5_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[4 ] = ab_pow4_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[3 ] = ab_pow3_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[2 ] = abb_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[1 ] = ab_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
assign ab_pow_im[0 ] = a_im[coef_width-1 : coef_width-ab_pow_width];//+ab_pow7_im[coef_width-ab_pow_width-1];
`endif 

wire	signed	[temp_var_width-1  :0]   x_re [0:15];

wire	signed	[temp_var_width+3  :0]   v_re;
reg	signed	[temp_var_width+3  :0]   v1_re;

wire    signed	[temp_var_width+3  :0]   y_re;
wire    signed	[temp_var_width+3  :0]   y1_re;

wire 	signed	[data_out_width-1:0]   y_re_trunc;

`ifdef COMPLEX
        wire	signed	[temp_var_width-1  :0]   x_im [0:15];

        wire	signed	[temp_var_width+3  :0]   v_im;
        reg	signed	[temp_var_width+3  :0]   v1_im;

        wire    signed	[temp_var_width+3  :0]   y_im;
        wire    signed	[temp_var_width+3  :0]   y1_im;

        wire 	signed	[data_out_width-1:0]   y_im_trunc;
`endif 
// x[0] = (dinp0 * a_re) delay M 	    = a*x(8n+8)
// x[1] = (dinp1 * ab_re) delay M  	    = a*b*x(8n+7)
// x[2] = (dinp2 * abb_re) delay M  	= a*b^2*x(8n+6)
// x[3] = (dinp3 * ab_pow3_re) delay M  = a*b^3*x(8n+5)
// x[4] = (dinp4 * ab_pow4_re) delay M  = a*b^4*x(8n+4)
// x[5] = (dinp5 * ab_pow5_re) delay M  = a*b^5*x(8n+3)
// x[6] = (dinp6 * ab_pow6_re) delay M  = a*b^6*x(8n+2)
// x[7] = (dinp7 * ab_pow7_re) delay M  = a*b^7*x(8n+1)

genvar i;
generate
    `ifdef COMPLEX
       for (i = 0; i < 16; i = i + 1) begin: mult_c_inst
            mult_x #(
                .A_width	(data_in_width	),
                .C_width	(coef_width	),
                .D_width	(coef_width	),
                .o_width	(temp_var_width	)
            ) inst_c (
                .clk    	(clk		 ),
                .rstn   	(rstn		 ),
                .en     	(en		     ),
                .a      	(dinp[i]	),         
                .c      	(ab_pow_re[i]	),
                .d      	(ab_pow_im[i]	),
                .Re     	(x_re[i]	),        
                .Im     	(x_im[i]	)
            );
        end
    `else
        for (i = 0; i < 16; i = i + 1) begin: mult_c_inst
            mult_real #(
                .A_width	(data_in_width	),
                .C_width	(coef_width	),
                .o_width	(temp_var_width	)
            ) inst_c (
                .clk    	(clk		 ),
                .rstn   	(rstn		 ),
                .en     	(en		     ),
                .din      	(dinp[i]	 ),         
                .coef     	(ab_pow_re[i]),
                .dout     	(x_re[i]	 )        
            );
        end
    `endif   
endgenerate


// v1 = sum_{i=0,1,...,7}{x[i]} delay M = sum_{i=0,1,...,7}{a*b^i*x(8n-i)}
assign	v_re       =   x_re[0] + x_re[1] +x_re[2] +x_re[3] +x_re[4] +x_re[5] +x_re[6] +x_re[7] + x_re[8] + x_re[9] + x_re[10] + x_re[11] + x_re[12] + x_re[13] + x_re[14] + x_re[15] ;

`ifdef COMPLEX
assign	v_im       =   x_im[0] + x_im[1] +x_im[2] +x_im[3] +x_im[4] +x_im[5] +x_im[6] +x_im[7] + x_im[8] + x_im[9] + x_im[10] + x_im[11] + x_im[12] + x_im[13] + x_im[14] + x_im[15] ;
`endif 

always @(posedge clk or negedge rstn) 
begin
    if (!rstn)
         begin
                v1_re    <=    'h0;
                `ifdef COMPLEX
                v1_im    <=    'h0;
                `endif
         end 
     else if(en)
         begin
                v1_re    <=    v_re;
                `ifdef COMPLEX
                v1_im    <=    v_im;
                `endif
          end
     else
         begin
                v1_re    <=    v1_re;
                `ifdef COMPLEX
                v1_im    <=    v1_im;
                `endif
          end
end

// y1 = (b^8 * y) delay M = b^8*y(8n-8)
// y = v1 + y1 = sum_{i=0,1,...,7}{a*b^i*x(8n-i)} + b^8*y(8n-8) = y(8n)
mult_real
#(
	 .A_width	(temp_var_width+4	) 
	,.C_width	(b_pow16_width		) 
	,.o_width	(temp_var_width+4 	) 
)
inst_c17 (
         .clk		(clk    	),
         .rstn		(rstn   	),
         .en		(en     	),
         .din		(y_re		),
         .coef		(b_pow16_re[coef_width-1 : coef_width-b_pow16_width]	),
         .dout		(y1_re     	)
);
assign	y_re       =   v1_re + y1_re;
`ifdef COMPLEX
assign	y_im       =   v1_im + y1_im;
`endif 

// dout = round(y) delay M = round(y(8n-8))
trunc #(
	 .diw	(temp_var_width+4		)
	,.msb	(temp_var_width-1		)
	,.lsb	(temp_var_width-data_out_width	)
) round_u1 (clk, rstn, en, y_re, y_re_trunc);
`ifdef COMPLEX
        trunc #(
             .diw	(temp_var_width+4		)
            ,.msb	(temp_var_width-1		)
            ,.lsb	(temp_var_width-data_out_width	)
        ) round_u2 (clk, rstn, en, y_im, y_im_trunc);
`endif 

assign	dout_re = y_re_trunc;
`ifdef COMPLEX
assign	dout_im = y_im_trunc;
`endif 

endmodule

